#include "rawSuperResolution.h"

rawDepthEstimation::rawDepthEstimation(int m, int M, const int BlockSize)
{
	mind = m;
	maxd = M;
	blocksize = BlockSize;
}

rawDepthEstimation::~rawDepthEstimation()
{

}

void rawDepthEstimation::estimate_rawdepth(const Mat& High, const Mat* View, const Mat* Mask, Mat& Depth)
{
	search_rawdisparity(High, View, Mask, Depth);
	save_depth(Depth);
}

void rawDepthEstimation::save_depth(const Mat& Depth)
{
	Mat depthmap = (Depth - (float)mind * (float)xscale) / ((float)(maxd - mind) * (float)xscale) * 255.0f;
	char str[64], str2[64];
	strcpy(str, FOLDER);
	sprintf(str2, "/bigSR/view%03d/depthmapbig.pgm", FRAME);
	strcat(str, str2);
	imwrite(str, depthmap);
}

// compute disparity using mask
void rawDepthEstimation::search_rawdisparity(const Mat& High, const Mat* View, const Mat* Mask, Mat& Depth)
{
	int l;
	int layer = (int)((maxd - mind) * cvRound(xscale) * 4.0 + 1.0);
	Mat blurHigh;
	PSF(High, blurHigh, xscale, yscale);
	float maxerror = 1000000.0f;	
	Mat minssd = Mat::ones(high_y, high_x, CV_32FC1) * maxerror;	
	Mat disparity = Mat::zeros(high_y, high_x, CV_32FC1);
	float d = (float)mind * (float)xscale;							// minimum depth

	// create cost maps from minimum disparity to max disparity
	for (l = 0; d < (float)maxd * (float)xscale + 0.25f; l += 1, d += 0.25f){
		Mat ssd = Mat::zeros(high_y, high_x, CV_32FC1);
		Mat count = Mat::zeros(high_y, high_x, CV_32FC1);			// record the number of times that stores a error

		for (int n = 0; n < FRAME; n++){
			if (n == (FRAME - 1) / 2){
				n = (FRAME - 1) / 2 + 1;
			}
			int move_y;
			int move_x;
			// amount of the movement depending on views
			if (FRAME == 25){
				move_y = 2 - (int)(n / 5);
				move_x = 2 - n % 5;
			}
			else if (FRAME == 49){
				move_y = 3 - (int)(n / 7);
				move_x = 3 - n % 7;
			}
			else if (FRAME == 81){
				move_y = 4 - (int)(n / 9);
				move_x = 4 - n % 9;
			}
			else if (FRAME == 121){
				move_y = 5 - (int)(n / 11);
				move_x = 5 - n % 11;
			}
			else if (FRAME == 169){
				move_y = 6 - (int)(n / 13);
				move_x = 6 - n % 13;
			}

			Mat largeMask, largeView;
			NNresize(View[n], Mask[n], largeView, largeMask);

			float *ssdPtr = (float *)(ssd.ptr<float>(0));
			float *countPtr = (float *)(count.ptr<float>(0));
			Vec3f *blurHighPtr = (Vec3f *)(blurHigh.ptr<Vec3f>(0));

			for (int j = 0; j < high_y; j++){
				for (int i = 0; i < high_x; i++){
					// ?
					int y = j + cvRound((double)move_y * (double)d / root3 * yscale / xscale);
					int x = i + cvRound((double)move_x * (double)d);
					if (y >= 0 && y < high_y && x >= 0 && x < high_x){
						int RGB = largeMask.at<unsigned char>(y, x);
						if (RGB != 3){
							float gaso = (*blurHighPtr)[RGB] - largeView.at<Vec3f>(y, x)[RGB];
							*ssdPtr += gaso * gaso;
							(*countPtr)++;
						}
					}
					ssdPtr++;
					countPtr++;
					blurHighPtr++;
				}
			}
		}

		float *countPtr = (float *)(count.ptr<float>(0));
		float *ssdPtr = (float *)(ssd.ptr<float>(0));

		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				if (*countPtr != 0.0f){
					*ssdPtr = *ssdPtr / *countPtr;
				}
				else{
					*ssdPtr = maxerror * 10.0f;
				}
				ssdPtr++;
				countPtr++;
			}
		}

		blur_ssd(ssd, maxerror);

		float *disparityPtr = (float *)(disparity.ptr<float>(0));
		float *minssdPtr = (float *)(minssd.ptr<float>(0));
		float *ssdMinimumPtr = (float *)(ssd.ptr<float>(0));

		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				if (*minssdPtr > *ssdMinimumPtr){
					*minssdPtr = *ssdMinimumPtr;
					*disparityPtr = d;
				}
				disparityPtr++;
				minssdPtr++;
				ssdMinimumPtr++;
			}
		}
	}

	medianBlur(disparity, Depth, 5);

	return;
}

void rawDepthEstimation::NNresize(const Mat& View, const Mat& Mask, Mat& LargeView, Mat& LargeMask)
{
	LargeView = Mat::zeros(high_y, high_x, CV_32FC3);
	LargeMask = Mat::ones(high_y, high_x, CV_8UC1) * 3;

	Vec3f *LargeViewPtr = (Vec3f *)(LargeView.ptr<Vec3f>(0));
	unsigned char *LargeMaskPtr = (unsigned char *)(LargeMask.ptr<unsigned char>(0));
	for (int j = 0; j < high_y; j++){
		for (int i = 0; i < high_x; i++){
			double y = (double)(j) / yscale;
			double x = (double)(i) / xscale;

			int m = cvRound(y);
			int nn = cvRound(x);

			if (m < low_y && nn < low_x){
				unsigned char RGB = Mask.at<unsigned char>(m, nn);
				*LargeMaskPtr = RGB;
				if (RGB != 3){
					(*LargeViewPtr)[RGB] = (float)View.at<Vec3b>(m, nn)[RGB];
				}
			}
			LargeViewPtr++;
			LargeMaskPtr++;
		}
	}

	return;
}

void rawDepthEstimation::blur_ssd(Mat& SSD, const float Maxerror)
{
	int margin = blocksize / 2;
	Mat cost = Mat::zeros(SSD.rows, SSD.cols, CV_32FC1);

	for (int j = margin; j < SSD.rows - margin; j++){
		float * costPtr = (float *)(&cost.ptr<float>(j)[margin]);
		for (int i = margin; i < SSD.cols - margin; i++){
			int count = 0;
			for (int s = -margin; s < margin + 1; s++){
				for (int t = -margin; t < margin + 1; t++){
					float value = SSD.at<float>(j + s, i + t);
					if (value < Maxerror * 10.0f - 10.0f){
						*costPtr += value;
						count++;
					}
				}
			}
			if (count != 0){
				*costPtr = *costPtr / (float)count;
			}
			costPtr++;
		}
	}

	SSD = cost.clone();

	return;
}

