#include "rawSuperResolution.h"

rawSuperResolution::rawSuperResolution(int iteration_, int iterationx_, double lambda_, double rho_)
{
	iteration = iteration_;
	iterationx = iterationx_;
	lambda = lambda_;
	rho = rho_;
}

rawSuperResolution::~rawSuperResolution()
{

}

void rawSuperResolution::raw_super_resolution(Mat& High, const Mat* View, const Mat* Mask, Mat& Disparity, bool joint_flag, bool yuv_flag)
{
	raw_increase_resolutionL1(High, View, Mask, Disparity, joint_flag, yuv_flag);
}

void rawSuperResolution::raw_increase_resolutionL1(Mat& High, const Mat* View, const Mat* Mask, Mat& Disparity, bool joint_flag, bool yuv_flag)
{
	int high_size = high_y * high_x;

	Mat z, w;
	if (yuv_flag == true){
		Mat fz;
		rgb2yuv(High, fz, 0);
		product_matrixD(fz, z, 0);
		Mat fw;
		rgb2yuv(High, fw, 0);
		product_matrixD(fw, w, 1);
	}
	else if (yuv_flag == false){
		product_matrixD(High, z, 0);
		product_matrixD(High, w, 1);
	}
	Mat u(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
	Mat rr(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));

	for (int ite = 0; ite < iteration; ite++){
		cout << "iteration : " << ite << endl;
		for (int x_ite = 0; x_ite < iterationx; x_ite++){
			cout << x_ite << endl;
		
			Mat sumerror = Mat::zeros(low_y, low_x, CV_32FC3);
			calculate_rawerror(High, View, Mask, Disparity, sumerror);
			double norm_error = norm(sumerror, NORM_L2);
			cout << "norm error = " << norm_error << endl;
			sumerror.release();
			Mat gradientmat(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			calculate_rawgradient_all(High, gradientmat, View, Mask, Disparity);
			
			Mat Dx(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			Mat Fx(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			if (yuv_flag == true){
				rgb2yuv(High, Fx, 0);
				product_matrixD(Fx, Dx, 0);
			}
			else if (yuv_flag == false){
				product_matrixD(High, Dx, 0);
			}
			Mat res(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			res = z - Dx + u;
			double norm_L1x = norm(Dx, NORM_L1);
			double norm_resX = norm(res, NORM_L2);
			Mat DtDx(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			product_matrixDt(res, DtDx, 0);
			Mat FtFx(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			if (yuv_flag == true){
				rgb2yuv(DtDx, FtFx, 1);
				gradientmat += (float)rho * FtFx;
			}
			else if (yuv_flag == false){
				gradientmat += (float)rho * DtDx;
			}
			Dx = Mat::zeros(high_y, high_x, CV_32FC3);
			if (yuv_flag == true){
				Fx = Mat::zeros(high_y, high_x, CV_32FC3);
				rgb2yuv(High, Fx, 0);
				product_matrixD(Fx, Dx, 1);
			}
			else if (yuv_flag == false){
				product_matrixD(High, Dx, 1);
			}
			res = Mat::zeros(high_y, high_x, CV_32FC3);
			res = w - Dx + rr;
			double norm_L1y = norm(Dx, NORM_L1);
			double norm_resY = norm(res, NORM_L2);
			DtDx = Mat::zeros(high_y, high_x, CV_32FC3);
			product_matrixDt(res, DtDx, 1);
			if (yuv_flag == true){
				FtFx = Mat::zeros(high_y, high_x, CV_32FC3);
				rgb2yuv(DtDx, FtFx, 1);
				gradientmat += (float)rho * FtFx;
			}
			else if (yuv_flag == false){
				gradientmat += (float)rho * DtDx;
			}
			gradientmat = -1.0f * gradientmat;

			Mat secondgradientmat(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			second_order_differential(Mask, Disparity, gradientmat, secondgradientmat);

			Mat Agradient(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
			Vec3f *gradientPtr = gradientmat.ptr<Vec3f>(0);
			Vec3f *secondPtr = secondgradientmat.ptr<Vec3f>(0);
			Vec3f *AgradientPtr = Agradient.ptr<Vec3f>(0);
			double norm_AtA = 0.0;
			for (int j = 0; j < high_y; j++){
				for (int i = 0; i < high_x; i++){
					for (int k = 0; k < 3; k++){
						(*AgradientPtr)[k] = (*gradientPtr)[k] * (*secondPtr)[k];
						norm_AtA += (double)fabs((*AgradientPtr)[k]);
					}
					gradientPtr++;
					secondPtr++;
					AgradientPtr++;
				}
			}
			Fx = Mat::zeros(high_y, high_x, CV_32FC3);
			Dx = Mat::zeros(high_y, high_x, CV_32FC3);
			DtDx = Mat::zeros(high_y, high_x, CV_32FC3);
			Agradient = Mat::zeros(high_y, high_x, CV_32FC3);
			secondgradientmat = Mat::zeros(high_y, high_x, CV_32FC3);
			if (yuv_flag == true){
				rgb2yuv(gradientmat, Fx, 0);
				product_matrixD(Fx, Dx, 0);
				product_matrixDt(Dx, DtDx, 0);
				rgb2yuv(DtDx, secondgradientmat, 1);
			}
			else if (yuv_flag == false){
				product_matrixD(gradientmat, Dx, 0);
				product_matrixDt(Dx, secondgradientmat, 0);
			}
			gradientPtr = gradientmat.ptr<Vec3f>(0);
			secondPtr = secondgradientmat.ptr<Vec3f>(0);
			AgradientPtr = Agradient.ptr<Vec3f>(0);
			double norm_Lx = 0.0;
			for (int j = 0; j < high_y; j++){
				for (int i = 0; i < high_x; i++){
					for (int k = 0; k < 3; k++){
						(*AgradientPtr)[k] = (*gradientPtr)[k] * (*secondPtr)[k];
						norm_Lx += (double)fabs((*AgradientPtr)[k]);
					}
					AgradientPtr++;
					gradientPtr++;
					secondPtr++;
				}
			}

			Fx = Mat::zeros(high_y, high_x, CV_32FC3);
			Dx = Mat::zeros(high_y, high_x, CV_32FC3);
			DtDx = Mat::zeros(high_y, high_x, CV_32FC3);
			Agradient = Mat::zeros(high_y, high_x, CV_32FC3);
			secondgradientmat = Mat::zeros(high_y, high_x, CV_32FC3);
			if (yuv_flag == true){
				rgb2yuv(gradientmat, Fx, 0);
				product_matrixD(Fx, Dx, 1);
				product_matrixDt(Dx, DtDx, 1);
				rgb2yuv(DtDx, secondgradientmat, 1);
			}
			else if (yuv_flag == false){
				product_matrixD(gradientmat, Dx, 1);
				product_matrixDt(Dx, secondgradientmat, 1);
			}
			gradientPtr = gradientmat.ptr<Vec3f>(0);
			secondPtr = secondgradientmat.ptr<Vec3f>(0);
			AgradientPtr = Agradient.ptr<Vec3f>(0);
			double norm_Ly = 0.0;
			for (int j = 0; j < high_y; j++){
				for (int i = 0; i < high_x; i++){
					for (int k = 0; k < 3; k++){
						(*AgradientPtr)[k] = (*gradientPtr)[k] * (*secondPtr)[k];
						norm_Ly += (double)fabs((*AgradientPtr)[k]);
					}
					AgradientPtr++;
					gradientPtr++;
					secondPtr++;
				}
			}

			double norm_gradientL2 = norm(gradientmat, NORM_L2);
			double norm_gradientL1 = norm(gradientmat, NORM_L1);
			double alpha = norm_gradientL2 * norm_gradientL2 / (norm_AtA + rho * norm_Lx + rho * norm_Ly);
			double L_SR = (norm_error * norm_error + rho * norm_resX * norm_resX + rho * norm_resY * norm_resY) / 2.0;
			double E_SR = (norm_error * norm_error) / 2.0 + lambda * norm_L1x + lambda * norm_L1y;
			cout << "alpha = " << alpha << endl;
			cout << "L_SR = " << L_SR << endl;
			if (x_ite == iterationx - 1){
				cout << "E_SR = " << E_SR << endl;
			}

			char str[64], str2[64];
			strcpy(str, FOLDER);
			if (joint_flag == false){
				sprintf(str2, "bigSR/view%03d/L1/%.1lf/normerror.txt", FRAME, lambda);
			}
			else if (joint_flag == true){
				sprintf(str2, "bigSR/view%03d/L1/%.1lf/joint/normerror.txt", FRAME, lambda);
			}
			strcat(str, str2);
			std::ofstream of(str, std::ios_base::app);
			if (ite == 0){
				if (x_ite == 0){
					of << "--------- norm error ------------" << std::endl;
				}
			}
			if (x_ite == 0){
				of << "ite" << "\t" << ite << std::endl;
			}
			of << x_ite << "\t" << norm_error  << "\t\t" << "E_SR" << "\t" << E_SR << "\t" << "Regular" << "\t\t" << lambda * norm_L1x + lambda * norm_L1y << std::endl;
			double t = 0.01;
			if (t > alpha * norm_gradientL1 / (double)high_size){
				break;
			}

			Vec3f *highPtr = (Vec3f *)(High.ptr<Vec3f>(0));
			gradientPtr = (Vec3f *)(gradientmat.ptr<Vec3f>(0));

			for (int j = 0; j < high_y; j++){
				for (int i = 0; i < high_x; i++){
					for (int k = 0; k < 3; k++){
						(*highPtr)[k] -= (float)alpha * (*gradientPtr)[k];
						if ((*highPtr)[k] < 0.0f){
							(*highPtr)[k] = 0.0f;
						}
						else if ((*highPtr)[k] > 255.0f){
							(*highPtr)[k] = 255.0f;
						}
					}
					highPtr++;
					gradientPtr++;
				}
			}
		}
		char str[64], str2[64];
		strcpy(str, FOLDER);
		if (joint_flag == false){
			sprintf(str2, "bigSR/view%03d/L1/%.1lf/sr%02d.ppm", FRAME, lambda, ite);
		}
		else if (joint_flag == true){
			sprintf(str2, "bigSR/view%03d/L1/%.1lf/joint/sr%02d.ppm", FRAME, lambda, ite);
		}
		strcat(str, str2);
		imwrite(str, High);
		if (joint_flag == true){
			int maxd = 1;
			int mind = -1;
			rawDepthEstimation *DE = new rawDepthEstimation(mind, maxd, 9);
			DE->estimate_rawdepth(High, View, Mask, Disparity);
			delete DE;
			strcpy(str, FOLDER);
			sprintf(str2, "bigSR/view%03d/L1/%.1lf/joint/depth/depth%02d.pgm", FRAME, lambda, ite);
			strcat(str, str2);
			Mat depthmap = (Disparity - (float)mind * (float)xscale) / ((float)(maxd - mind) * (float)xscale) * 255.0f;
			imwrite(str, depthmap);
		}
		Mat Fx(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
		Mat Dx(high_y, high_x, CV_32FC3, Scalar(0, 0, 0));
		if (yuv_flag == true){
			rgb2yuv(High, Fx, 0);
			product_matrixD(Fx, Dx, 0);
		}
		else if (yuv_flag == false){
			product_matrixD(High, Dx, 0);
		}
		Mat TVu;
		TVu = Dx - u;
		shrinkage(TVu, z);
		u = u + z - Dx;
		Dx = Mat::zeros(high_y, high_x, CV_32FC3);
		if (yuv_flag == true){
			Fx = Mat::zeros(high_y, high_x, CV_32FC3);
			rgb2yuv(High, Fx, 0);
			product_matrixD(Fx, Dx, 1);
		}
		else if (yuv_flag == false){
			product_matrixD(High, Dx, 1);
		}
		Mat TVr;
		TVr = Dx - rr;
		shrinkage(TVr, w);
		rr = rr + w - Dx;
	}

	return;
}

void rawSuperResolution::calculate_rawerror(Mat& High, const Mat* View, const Mat* Mask, const Mat& Disparity, Mat& Sumerror)
{
	Mat blurHigh;
	PSF(High, blurHigh, xscale, yscale);
	for (int n = 0; n < FRAME; n++){
		int move_y;
		int move_x;

		if (FRAME == 25){
			move_y = 2 - (int)(n / 5);
			move_x = 2 - n % 5;
		}
		else if (FRAME == 49){
			move_y = 3 - (int)(n / 7);
			move_x = 3 - n % 7;
		}
		else if (FRAME == 81){
			move_y = 4 - (int)(n / 9);
			move_x = 4 - n % 9;
		}
		else if (FRAME == 121){
			move_y = 5 - (int)(n / 11);
			move_x = 5 - n % 11;
		}
		else if (FRAME == 169){
			move_y = 6 - (int)(n / 13);
			move_x = 6 - n % 13;
		}
		Vec3f *blurHighPtr = (Vec3f *)(blurHigh.ptr<Vec3f>(0));
		float *motionPtr = (float *)(Disparity.ptr<float>(0));

		Mat motionHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Mat depth = Mat::ones(high_y, high_x, CV_32FC1) * -100.0f;

		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				int y = j + cvRound((double)move_y * (double)(*motionPtr) / root3 * yscale / xscale);
				int x = i + cvRound((double)move_x * (double)(*motionPtr));

				if (y >= 0 && y < high_y && x >= 0 && x < high_x){
					if (depth.at<float>(y, x) < *motionPtr){
						depth.at<float>(y, x) = *motionPtr;
						for (int k = 0; k < 3; k++){
							motionHigh.at<Vec3f>(y, x)[k] = (*blurHighPtr)[k];
						}
					}
				}
				motionPtr++;
				blurHighPtr++;
			}
		}
		
		Mat error = Mat::zeros(low_y, low_x, CV_32FC3);
		Vec3b *lowPtr = (Vec3b *)(View[n].ptr<Vec3b>(0));
		Vec3f *errorPtr = (Vec3f *)(error.ptr<Vec3f>(0));
		unsigned char *maskPtr = (unsigned char *)(Mask[n].ptr<unsigned char>(0));

		for (int j = 0; j < low_y; j++){
			for (int i = 0; i < low_x; i++){
				double y = (double)j*yscale;
				double x = (double)i*xscale;

				int m = (int)y;
				int nn = (int)x;

				float q = (float)y - (float)m;
				float p = (float)x - (float)nn;

				int RGB = *maskPtr;
				float b = 0.0f, g = 0.0f, r = 0.0f;
				if (m >= 0 && m < high_y - 1 && nn >= 0 && nn < high_x - 1){
					if ((int)depth.at<float>(m, nn) != -100 && (int)depth.at<float>(m, nn + 1) != -100 && (int)depth.at<float>(m + 1, nn) != -100 && (int)depth.at<float>(m + 1, nn + 1) != -100){
						b = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[0] + p*motionHigh.at<Vec3f>(m, nn + 1)[0])
							+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[0] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[0]);

						g = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[1] + p*motionHigh.at<Vec3f>(m, nn + 1)[1])
							+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[1] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[1]);

						r = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[2] + p*motionHigh.at<Vec3f>(m, nn + 1)[2])
							+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[2] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[2]);
					}
				}

				if ((b == 0 && g == 0 && r == 0) || RGB == 3){
					for (int k = 0; k < 3; k++){
						(*errorPtr)[k] = 0.0f;
					}
				}
				else{
					if (RGB == 0){
						(*errorPtr)[0] = (float)(*lowPtr)[0] - b;
					}
					else if (RGB == 1){
						(*errorPtr)[1] = (float)(*lowPtr)[1] - g;
					}
					else if (RGB == 2){
						(*errorPtr)[2] = (float)(*lowPtr)[2] - r;
					}
				}
			
				lowPtr++;
				errorPtr++;
				maskPtr++;
			}
		}
		
		Vec3f *errorHighPtr = error.ptr<Vec3f>(0);
		Vec3f *sumerrorPtr = Sumerror.ptr<Vec3f>(0);
		for (int j = 0; j < low_y; j++){
			for (int i = 0; i < low_x; i++){
				(*sumerrorPtr)[0] += fabs((*errorHighPtr)[0]);
				(*sumerrorPtr)[1] += fabs((*errorHighPtr)[1]);
				(*sumerrorPtr)[2] += fabs((*errorHighPtr)[2]);
				errorHighPtr++;
				sumerrorPtr++;
			}
		}
	}

	return;
}

void rawSuperResolution::second_order_differential(const Mat* Mask, const Mat& Disparity, Mat& Gradient, Mat& Square)
{
	Mat blurGradient;
	PSF(Gradient, blurGradient, xscale, yscale);
	for (int n = 0; n < FRAME; n++){
		int move_y;
		int move_x;

		if (FRAME == 25){
			move_y = 2 - (int)(n / 5);
			move_x = 2 - n % 5;
		}
		else if (FRAME == 49){
			move_y = 3 - (int)(n / 7);
			move_x = 3 - n % 7;
		}
		else if (FRAME == 81){
			move_y = 4 - (int)(n / 9);
			move_x = 4 - n % 9;
		}
		else if (FRAME == 121){
			move_y = 5 - (int)(n / 11);
			move_x = 5 - n % 11;
		}
		else if (FRAME == 169){
			move_y = 6 - (int)(n / 13);
			move_x = 6 - n % 13;
		}
		Vec3f *blurHighPtr = (Vec3f *)(blurGradient.ptr<Vec3f>(0));
		float *motionPtr = (float *)(Disparity.ptr<float>(0));

		Mat motionHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Mat depth = Mat::ones(high_y, high_x, CV_32FC1) * -100.0f;
		
		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				int y = j + cvRound((double)move_y * (double)(*motionPtr) / root3 * yscale / xscale);
				int x = i + cvRound((double)move_x * (double)(*motionPtr));

				if (y >= 0 && y < high_y && x >= 0 && x < high_x){
					if (depth.at<float>(y, x) < *motionPtr){
						depth.at<float>(y, x) = *motionPtr;
						for (int k = 0; k < 3; k++){
							motionHigh.at<Vec3f>(y, x)[k] = (*blurHighPtr)[k];
						}
					}
				}
				motionPtr++;
				blurHighPtr++;
			}
		}
		
		Mat error = Mat::zeros(low_y, low_x, CV_32FC3);
		Vec3f *errorPtr = (Vec3f *)(error.ptr<Vec3f>(0));
		unsigned char *maskPtr = (unsigned char *)(Mask[n].ptr<unsigned char>(0));
	
		// downsampling
		for (int j = 0; j < low_y; j++){
			for (int i = 0; i < low_x; i++){
				double y = (double)j*yscale;
				double x = (double)i*xscale;

				int m = (int)y;
				int nn = (int)x;

				float q = (float)y - (float)m;
				float p = (float)x - (float)nn;

				int RGB = *maskPtr;
				float b = 0.0f, g = 0.0f, r = 0.0f;
					if (m >= 0 && m < high_y - 1 && nn >= 0 && nn < high_x - 1){
						if ((int)depth.at<float>(m, nn) != -100 && (int)depth.at<float>(m, nn + 1) != -100 && (int)depth.at<float>(m + 1, nn) != -100 && (int)depth.at<float>(m + 1, nn + 1) != -100){
							b = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[0] + p*motionHigh.at<Vec3f>(m, nn + 1)[0])
								+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[0] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[0]);

							g = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[1] + p*motionHigh.at<Vec3f>(m, nn + 1)[1])
								+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[1] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[1]);

							r = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[2] + p*motionHigh.at<Vec3f>(m, nn + 1)[2])
								+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[2] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[2]);
						}
					}

				if (RGB == 3){
					for (int k = 0; k < 3; k++){
						(*errorPtr)[k] = 0.0f;
					}
				}
				else{
					if (RGB == 0){
						(*errorPtr)[0] = b;
					}
					else if (RGB == 1){
						(*errorPtr)[1] = g;
					}
					else if (RGB == 2){
						(*errorPtr)[2] = r;
					}
				}
				errorPtr++;
				maskPtr++;
			}
		}

		Mat upsamplingHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Vec3f *AImgPtr = (Vec3f *)(error.ptr<Vec3f>(0));
		unsigned char *maskHighPtr = (unsigned char *)(Mask[n].ptr<unsigned char>(0));
		
		for (int j = 0; j < low_y; j++){
			for (int i = 0; i < low_x; i++){
				double y = (double)j*yscale;
				double x = (double)i*xscale;

				int m = (int)y;
				int nn = (int)x;

				float q = (float)y - (float)m;
				float p = (float)x - (float)nn;

				int RGB = *maskHighPtr;
				if (m >= 0 && m < high_y - 1 && nn >= 0 && nn < high_x - 1){
					if (RGB != 3){
						upsamplingHigh.at<Vec3f>(m, nn)[RGB] += (1.0f - q)*(1.0f - p)*(*AImgPtr)[RGB];
						upsamplingHigh.at<Vec3f>(m, nn + 1)[RGB] += (1.0f - q)*p*(*AImgPtr)[RGB];
						upsamplingHigh.at<Vec3f>(m + 1, nn)[RGB] += q*(1.0f - p)*(*AImgPtr)[RGB];
						upsamplingHigh.at<Vec3f>(m + 1, nn + 1)[RGB] += q*p*(*AImgPtr)[RGB];
					}
				}
				maskHighPtr++;
				AImgPtr++;
			}
		}

		if (FRAME == 25){
			move_y = -2 + (int)(n / 5);
			move_x = -2 + n % 5;
		}
		else if (FRAME == 49){
			move_y = -3 + (int)(n / 7);
			move_x = -3 + n % 7;
		}
		else if (FRAME == 81){
			move_y = -4 + (int)(n / 9);
			move_x = -4 + n % 9;
		}
		else if (FRAME == 121){
			move_y = -5 + (int)(n / 11);
			move_x = -5 + n % 11;
		}
		else if (FRAME == 169){
			move_y = -6 + (int)(n / 13);
			move_x = -6 + n % 13;
		}

		Mat reverseMotionHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Vec3f *upsamplingHighPtr = (Vec3f *)(upsamplingHigh.ptr<Vec3f>(0));
		float *depthPtr = (float *)(depth.ptr<float>(0));

		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				int y = j + cvRound((double)move_y * (double)(*depthPtr) / root3 * yscale / xscale);
				int x = i + cvRound((double)move_x * (double)(*depthPtr));
				if (y >= 0 && y < high_y && x >= 0 && x < high_x){
					if ((int)(*depthPtr) != -100){
						for (int k = 0; k < 3; k++){
							reverseMotionHigh.at<Vec3f>(y, x)[k] = (*upsamplingHighPtr)[k];
						}
					}
				}
				upsamplingHighPtr++;
				depthPtr++;
			}
		}

		Mat reverseBlurHigh;
		PSF(reverseMotionHigh, reverseBlurHigh, xscale, yscale);

		Square += reverseBlurHigh;
	}

	return;
}

void rawSuperResolution::calculate_rawgradient_all(Mat& High, Mat& Gradient, const Mat* View, const Mat* Mask, const Mat& Disparity)
{
	Mat blurHigh;
	PSF(High, blurHigh, xscale, yscale);
	char str[64], str2[64];
	strcpy(str, FOLDER);
	sprintf(str2, "bigSR/view%03d/blur.ppm", FRAME);
	strcat(str, str2);
	imwrite(str, blurHigh);
	for (int n = 0; n < FRAME; n++){
		int move_y;
		int move_x;

		if (FRAME == 25){
			move_y = 2 - (int)(n / 5);
			move_x = 2 - n % 5;
		}
		else if (FRAME == 49){
			move_y = 3 - (int)(n / 7);
			move_x = 3 - n % 7;
		}
		else if (FRAME == 81){
			move_y = 4 - (int)(n / 9);
			move_x = 4 - n % 9;
		}
		else if (FRAME == 121){
			move_y = 5 - (int)(n / 11);
			move_x = 5 - n % 11;
		}
		else if (FRAME == 169){
			move_y = 6 - (int)(n / 13);
			move_x = 6 - n % 13;
		}
		Vec3f *blurHighPtr = (Vec3f *)(blurHigh.ptr<Vec3f>(0));
		float *motionPtr = (float *)(Disparity.ptr<float>(0));

		Mat motionHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Mat depth = Mat::ones(high_y, high_x, CV_32FC1) * -100.0f;

		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				int y = j + cvRound((double)move_y * (double)(*motionPtr) / root3 * yscale / xscale);
				int x = i + cvRound((double)move_x * (double)(*motionPtr));

				if (y >= 0 && y < high_y && x >= 0 && x < high_x){
					if (depth.at<float>(y, x) < *motionPtr){
						depth.at<float>(y, x) = *motionPtr;
						for (int k = 0; k < 3; k++){
							motionHigh.at<Vec3f>(y, x)[k] = (*blurHighPtr)[k];
						}
					}
				}
				motionPtr++;
				blurHighPtr++;
			}
		}
		if (n == 10){
			strcpy(str, FOLDER);
			sprintf(str2, "bigSR/view%03d/motion10.ppm", FRAME);
			strcat(str, str2);
			imwrite(str, motionHigh);
		}
		if (n == 60){
			strcpy(str, FOLDER);
			sprintf(str2, "bigSR/view%03d/motion60.ppm", FRAME);
			strcat(str, str2);
			imwrite(str, motionHigh);
		}
		Mat error = Mat::zeros(low_y, low_x, CV_32FC3);
		Vec3b *lowPtr = (Vec3b *)(View[n].ptr<Vec3b>(0));
		Vec3f *errorPtr = (Vec3f *)(error.ptr<Vec3f>(0));
		unsigned char *maskPtr = (unsigned char *)(Mask[n].ptr<unsigned char>(0));

		for (int j = 0; j < low_y; j++){
			for (int i = 0; i < low_x; i++){
				double y = (double)j*yscale;
				double x = (double)i*xscale;

				int m = (int)y;
				int nn = (int)x;

				float q = (float)y - (float)m;
				float p = (float)x - (float)nn;

				int RGB = *maskPtr;
				float b = 0.0f, g = 0.0f, r = 0.0f;
				if (m >= 0 && m < high_y - 1 && nn >= 0 && nn < high_x - 1){
					if ((int)depth.at<float>(m, nn) != -100 && (int)depth.at<float>(m, nn + 1) != -100 && (int)depth.at<float>(m + 1, nn) != -100 && (int)depth.at<float>(m + 1, nn + 1) != -100){
						b = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[0] + p*motionHigh.at<Vec3f>(m, nn + 1)[0])
							+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[0] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[0]);

						g = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[1] + p*motionHigh.at<Vec3f>(m, nn + 1)[1])
							+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[1] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[1]);

						r = (1.0f - q)*((1.0f - p)*motionHigh.at<Vec3f>(m, nn)[2] + p*motionHigh.at<Vec3f>(m, nn + 1)[2])
							+ q*((1.0f - p)*motionHigh.at<Vec3f>(m + 1, nn)[2] + p*motionHigh.at<Vec3f>(m + 1, nn + 1)[2]);
					}
				}

				if ((b == 0 && g == 0 && r == 0) || RGB == 3){
					for (int k = 0; k < 3; k++){
						(*errorPtr)[k] = 0.0f;
					}
				}
				else{
					if (RGB == 0){
						(*errorPtr)[0] = (float)(*lowPtr)[0] - b;
					}
					else if (RGB == 1){
						(*errorPtr)[1] = (float)(*lowPtr)[1] - g;
					}
					else if (RGB == 2){
						(*errorPtr)[2] = (float)(*lowPtr)[2] - r;
					}
				}
				
				lowPtr++;
				errorPtr++;
				maskPtr++;
			}
		}
		if (n == 10){
			strcpy(str, FOLDER);
			sprintf(str2, "bigSR/view%03d/error10.ppm", FRAME);
			strcat(str, str2);
			imwrite(str, error);
		}
		if (n == 60){
			strcpy(str, FOLDER);
			sprintf(str2, "bigSR/view%03d/error60.ppm", FRAME);
			strcat(str, str2);
			imwrite(str, error);
		}

		Mat upsamplingHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Vec3f *AImgPtr = (Vec3f *)(error.ptr<Vec3f>(0));
		unsigned char *maskHighPtr = (unsigned char *)(Mask[n].ptr<unsigned char>(0));

		for (int j = 0; j < low_y; j++){
			for (int i = 0; i < low_x; i++){
				double y = (double)j*yscale;
				double x = (double)i*xscale;

				int m = (int)y;
				int nn = (int)x;

				float q = (float)y - (float)m;
				float p = (float)x - (float)nn;

				int RGB = *maskHighPtr;
				if (m >= 0 && m < high_y - 1 && nn >= 0 && nn < high_x - 1){
					if (RGB != 3){
						upsamplingHigh.at<Vec3f>(m, nn)[RGB] += (1.0f - q)*(1.0f - p)*(*AImgPtr)[RGB];
						upsamplingHigh.at<Vec3f>(m, nn + 1)[RGB] += (1.0f - q)*p*(*AImgPtr)[RGB];
						upsamplingHigh.at<Vec3f>(m + 1, nn)[RGB] += q*(1.0f - p)*(*AImgPtr)[RGB];
						upsamplingHigh.at<Vec3f>(m + 1, nn + 1)[RGB] += q*p*(*AImgPtr)[RGB];
					}
				}
				maskHighPtr++;
				AImgPtr++;
			}
		}

		if (FRAME == 25){
			move_y = -2 + (int)(n / 5);
			move_x = -2 + n % 5;
		}
		else if (FRAME == 49){
			move_y = -3 + (int)(n / 7);
			move_x = -3 + n % 7;
		}
		else if (FRAME == 81){
			move_y = -4 + (int)(n / 9);
			move_x = -4 + n % 9;
		}
		else if (FRAME == 121){
			move_y = -5 + (int)(n / 11);
			move_x =-5 + n % 11;
		}
		else if (FRAME == 169){
			move_y = -6 + (int)(n / 13);
			move_x = -6 + n % 13;
		}

		Mat reverseMotionHigh = Mat::zeros(high_y, high_x, CV_32FC3);
		Vec3f *upsamplingHighPtr = (Vec3f *)(upsamplingHigh.ptr<Vec3f>(0));
		float *depthPtr = (float *)(depth.ptr<float>(0));

		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				int y = j + cvRound((double)move_y * (double)(*depthPtr) / root3 * yscale / xscale);
				int x = i + cvRound((double)move_x * (double)(*depthPtr));
				if (y >= 0 && y < high_y && x >= 0 && x < high_x){
					if ((int)(*depthPtr) != -100){
						for (int k = 0; k < 3; k++){
							reverseMotionHigh.at<Vec3f>(y, x)[k] = (*upsamplingHighPtr)[k];
						}
					}
				}
				upsamplingHighPtr++;
				depthPtr++;
			}
		}

		Mat reverseBlurHigh;
		PSF(reverseMotionHigh, reverseBlurHigh, xscale, yscale);

		Gradient += reverseBlurHigh;
	}

	return;
}

void rawSuperResolution::rgb2yuv(Mat& Img, Mat& OutImg, bool flag)
{
	Mat yuv = (Mat_<float>(3, 3) << 0.299f, 0.587f, 0.114f, -0.169f, -0.331f, 0.5f, 0.5f, -0.419f, 0.081f);
	OutImg = Mat::zeros(high_y, high_x, CV_32FC3);
	if (flag == false){
		Vec3f *OutImgPtr = OutImg.ptr<Vec3f>(0);
		Vec3f *ImgPtr = Img.ptr<Vec3f>(0);
		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				(*OutImgPtr)[0] = yuv.at<float>(0, 0) * (*ImgPtr)[0] + yuv.at<float>(0, 1) * (*ImgPtr)[1] + yuv.at<float>(0, 2) * (*ImgPtr)[2];
				(*OutImgPtr)[1] = yuv.at<float>(1, 0) * (*ImgPtr)[0] + yuv.at<float>(1, 1) * (*ImgPtr)[1] + yuv.at<float>(1, 2) * (*ImgPtr)[2];
				(*OutImgPtr)[2] = yuv.at<float>(2, 0) * (*ImgPtr)[0] + yuv.at<float>(2, 1) * (*ImgPtr)[1] + yuv.at<float>(2, 2) * (*ImgPtr)[2];
				ImgPtr++;
				OutImgPtr++;
			}
		}
	}
	else if (flag == true){
		Vec3f *OutImgPtr = OutImg.ptr<Vec3f>(0);
		Vec3f *ImgPtr = Img.ptr<Vec3f>(0);
		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				(*OutImgPtr)[0] = yuv.at<float>(0, 0) * (*ImgPtr)[0] + yuv.at<float>(1, 0) * (*ImgPtr)[1] + yuv.at<float>(2, 0) * (*ImgPtr)[2];
				(*OutImgPtr)[1] = yuv.at<float>(0, 1) * (*ImgPtr)[0] + yuv.at<float>(1, 1) * (*ImgPtr)[1] + yuv.at<float>(2, 1) * (*ImgPtr)[2];
				(*OutImgPtr)[2] = yuv.at<float>(0, 2) * (*ImgPtr)[0] + yuv.at<float>(1, 2) * (*ImgPtr)[1] + yuv.at<float>(2, 2) * (*ImgPtr)[2];
				ImgPtr++;
				OutImgPtr++;
			}
		}
	}

}

void rawSuperResolution::product_matrixD(Mat& Img, Mat& DImg, bool flag)
{
	float sobel[2] = { -1.0f, 1.0f };

	DImg = Mat::zeros(high_y, high_x, CV_32FC3);

	if (flag == false){						// x direction
		Vec3f *sobelPtr = (Vec3f *)(DImg.ptr<Vec3f>(0));
		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				for (int t = 0; t < 2; t++){
					if (i + t >= 0 && i + t < high_x){
						for (int k = 0; k < 3; k++){
							(*sobelPtr)[k] += sobel[t] * Img.at<Vec3f>(j, i + t)[k];
						}
					}
				}
				sobelPtr++;
			}
		}
	}
	else if (flag == true){					// y direction
		Mat sobel_T = DImg.t();	
		Mat gradient_T = Img.t();
		Vec3f *sobel_TPtr = (Vec3f *)(sobel_T.ptr<Vec3f>(0));
		for (int j = 0; j < gradient_T.rows; j++){
			for (int i = 0; i < gradient_T.cols; i++){
				for (int t = 0; t < 2; t++){
					if (i + t >= 0 && i + t < gradient_T.cols){
						for (int k = 0; k < 3; k++){
							(*sobel_TPtr)[k] += sobel[t] * gradient_T.at<Vec3f>(j, i + t)[k];
						}
					}
				}
				sobel_TPtr++;
			}
		}

		DImg = sobel_T.t();
	}

	return;
}

void rawSuperResolution::product_matrixDt(Mat& Img, Mat& DtImg, bool flag)
{
	float sobel[2] = { 1.0f, -1.0f };

	DtImg = Mat::zeros(high_y, high_x, CV_32FC3);

	if (flag == false){						// x direction
		Vec3f *sobelPtr = (Vec3f *)(DtImg.ptr<Vec3f>(0));
		for (int j = 0; j < high_y; j++){
			for (int i = 0; i < high_x; i++){
				for (int t = -1; t < 1; t++){
					if (i + t >= 0 && i + t < high_x){
						for (int k = 0; k < 3; k++){
							(*sobelPtr)[k] += sobel[t + 1] * Img.at<Vec3f>(j, i + t)[k];
						}
					}
				}
				sobelPtr++;
			}
		}
	}
	else if (flag == true){					// y direction
		Mat sobel_T = DtImg.t();
		Mat gradient_T = Img.t();
		Vec3f *sobel_TPtr = (Vec3f *)(sobel_T.ptr<Vec3f>(0));
		for (int j = 0; j < gradient_T.rows; j++){
			for (int i = 0; i < gradient_T.cols; i++){
				for (int t = -1; t < 1; t++){
					if (i + t >= 0 && i + t < gradient_T.cols){
						for (int k = 0; k < 3; k++){
							(*sobel_TPtr)[k] += sobel[t + 1] * gradient_T.at<Vec3f>(j, i + t)[k];
						}
					}
				}
				sobel_TPtr++;
			}
		}

		DtImg = sobel_T.t();
	}

	return;
}

void rawSuperResolution::product_matrixL(Mat& Img, Mat& LImg)
{
	Mat Laplacian = (Mat_<float>(3, 3) << 0.0f, -1.0f, 0.0f, -1.0f, 4.0f, -1.0f, 0.0f, -1.0f, 0.0f);
	LImg = Mat::zeros(high_y, high_x, CV_32FC3);
	Vec3f *LImgPtr = LImg.ptr<Vec3f>(0);
	for (int j = 0; j < high_y; j++){
		for (int i = 0; i < high_x; i++){
			float *kernelPtr = Laplacian.ptr<float>(0);
			for (int k = -1; k < 2; k++){
				for (int l = -1; l < 2; l++){
					if (j + k >= 0 && j + k < high_y && i + l >= 0 && i + l < high_x){
						for (int n = 0; n < 3; n++){
							(*LImgPtr)[n] += *kernelPtr *Img.at<Vec3f>(j + k, i + l)[n];
						}
					}
					kernelPtr++;
				}
			}
			LImgPtr++;
		}
	}

	return;
}

void rawSuperResolution::shrinkage(Mat& Img, Mat& Out)
{
	Out = Mat::zeros(high_y, high_x, CV_32FC3);
	Vec3f *ImgPtr = (Vec3f *)(Img.ptr<Vec3f>(0));
	Vec3f *OutPtr = (Vec3f *)(Out.ptr<Vec3f>(0));

	for (int j = 0; j < high_y; j++){
		for (int i = 0; i < high_x; i++){
			for (int k = 0; k < 3; k++){
				int sign = ((int)(*ImgPtr)[k] > 0) - ((int)(*ImgPtr)[k] < 0);
				(*OutPtr)[k] = (float)sign * (float)max(fabs((*ImgPtr)[k]) - lambda / rho, 0.0);
			}
			ImgPtr++;
			OutPtr++;
		}
	}

	return;
}
