#pragma once

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <algorithm>
#include <fstream>
#include <opencv2/opencv.hpp>

using namespace cv;
using namespace std;

#pragma warning(disable:4996)

#define FOLDER "./result/"

#define root3 1.73205080757
#define FRAME 81					// number of viewpoint
static const int high_y = 1634;		// number of pixels in vertical direction of high-resolution image
static const int high_x = 2450;		// number of pixels in horizontal direction of high-resolution image
static const int low_y = 433;		// number of pixels in vertical direction of low-resolution image
static const int low_x = 1084;		// number of pixels in horizontal direction of low-resolution image
static const double xscale = 2450.0 / 1084.0;	// magnification of horizontal direction
static const double yscale = 1634.0 / 433.0;	// magnification of vertical direction

// class which generates subaperture-images from a RAW image
class viewCreate{
private:
	Mat raw;					// RAW image (color)
	Mat gray;					// RAW image (gray)
	int view_y, view_x;			// image size of a subaperture-image
	double space_y, space_x;	// pixel spacing of a sub-image
	double slope_y, slope_x;	// slope of a RAW image
	double offset_y, offset_x;	// pixel position located center of micro-lens in the upper left corner  

	void interpolate_subaperture(const Mat& Input, Mat& Output);
	void extract_viewandmask(const Mat& Raw, Mat* View, Mat* Mask);
	void extract_center(const Mat& Raw, Mat& Center);
	void save_subaperture(const Mat* View);

public:
	viewCreate(const char* color, const char* bayer);
	~viewCreate();

	void create_mask(Mat* View, Mat* Mask, Mat& Center);
};

// class which estimates a high-resolution depth map
class rawDepthEstimation{
private:
	int mind, maxd;		// minimum, max parallax
	int blocksize;		// block size

	void search_rawdisparity(const Mat&High, const Mat* View, const Mat* Mask, Mat& Depth);
	void NNresize(const Mat& View, const Mat& Mask, Mat& LargeView, Mat& LargeMask);
	void blur_ssd(Mat& SSD, const float Maxerror);	
	void save_depth(const Mat& Depth);	

public:
	rawDepthEstimation(int m, int M, const int BlockSize);
	~rawDepthEstimation();

	void estimate_rawdepth(const Mat& High, const Mat* View, const Mat* Mask, Mat& Depth);
};

class rawSuperResolution{
private:
	int iteration;		// iteration number of ADMM
	int iterationx;		// iteration number of gradient decent method
	double lambda;		// regularization parameter
	double rho;

	void raw_increase_resolutionL1(Mat& High, const Mat* View, const Mat* Mask, Mat& Disparity, bool joint_flag, bool yuv_flag);
	void calculate_rawerror(Mat& High, const Mat* View, const Mat* Mask, const Mat& Disparity, Mat& Sumerror);
	void second_order_differential(const Mat* Mask, const Mat& Disparity, Mat& Gradient, Mat& Square);
	void calculate_rawgradient_all(Mat& High, Mat& Gradient, const Mat* View, const Mat* Mask, const Mat& Disparity);
	void rgb2yuv(Mat& Img, Mat& Yuv, bool flag);
	void product_matrixD(Mat& Img, Mat& DImg, bool flag);
	void product_matrixDt(Mat& Img, Mat& DtImg, bool flag);
	void product_matrixL(Mat& Img, Mat& LImg);
	void shrinkage(Mat& Img, Mat& Out);

public:
	rawSuperResolution(int iteration, int iterationx, double lambda, double rho);
	~rawSuperResolution();

	void raw_super_resolution(Mat& High, const Mat* View, const Mat* Mask, Mat& Disparity, bool joint_flag, bool yuv_flag);
};

void PSF(const Mat& input, Mat& output, double xscale, double yscale);

