#include "rawSuperResolution.h"

viewCreate::viewCreate(const char *color, const char* bayer)
{
	raw = imread(color, -1);
	gray = imread(bayer, -1);
	view_y = 433;
	view_x = 542;
	space_y = 12.37046396;		// obtained from LFP file
	space_x = 14.28239569;	
	double theta = 0.0008988660784;
	slope_y = space_x*tan(theta);
	slope_x = space_y*tan(-1.0 * theta);
	offset_y = 11.77484499;
	offset_x = 1.084513694;
}

viewCreate::~viewCreate()
{
	raw.release();
	gray.release();
}

// creating sub-aperture images by directmethod, masks of sub-aperture images, and a center viewpoint sub-aperture image 
void viewCreate::create_mask(Mat* View, Mat* Mask, Mat& Center)
{
	extract_viewandmask(gray, View, Mask);
	extract_center(raw, Center);
	//save_subaperture(View);

}

void viewCreate::save_subaperture(const Mat* View)
{
	for (int n = 0; n < FRAME; n++){
		char str[64], str2[64];
		strcpy(str, FOLDER);
		sprintf(str2, "view/view%03d.ppm", n);
		strcat(str, str2);
		imwrite(str, View[n]);
	}
}

// creating a center viewpoint image for initial magnified image
void viewCreate::extract_center(const Mat& Raw, Mat& Center)
{
	int X, Y;
	double x, y;

	// expanding double in x direction
	Mat subaperture = Mat::zeros(view_y, 2 * view_x, CV_32FC3);

	int s = -1;
	int t = 6;

	for (int j = 0; j < view_y; j = j + 2){
		for (int i = 0; i < view_x; i++){
			// pixel position of center viewpoint in each subimage
			y = offset_y + space_y * (double)j + slope_y * (double)i;
			x = offset_x + space_x * (double)i + slope_x * (double)j;

			// round the pixel position
			int xi = cvRound(x);
			int yi = cvRound(y);
			double xd = fabs(x - (double)xi);
			double yd = fabs(y - (double)yi);
			Y = yi + s;
			X = xi + t;

			if (Y >= 0 && Y < gray.rows - 1 && X >= 0 && X < gray.cols - 1){
				for (int k = 0; k < 3; k++){
					// in the case of even-numbered rows, a pixel from RAW image is stored in odd-numbered columns of the sub-aperture image
					subaperture.at<Vec3f>(j, 2 * i + 1)[k] = (float)Raw.at<Vec3b>(Y, X)[k];
				}
			}
		}
	}

	for (int j = 1; j < view_y; j = j + 2){
		for (int i = 0; i < view_x; i++){
			// pixel position of center viewpoint in each subimage
			y = offset_y + space_y * (double)j + slope_y * (double)i;
			x = offset_x - space_x / 2.0 + space_x * (double)i + slope_x * (double)j;

			// round the pixel position
			int xi = cvRound(x);
			int yi = cvRound(y);
			double xd = fabs(x - (double)xi);
			double yd = fabs(y - (double)yi);
			Y = yi + s;
			X = xi + t;
			
			if (Y >= 0 && Y < gray.rows - 1 && X >= 0 && X < gray.cols - 1){
				for (int k = 0; k < 3; k++){
					// in the case of odd-numbered rows, a pixel from RAW image is stored in even-numbered columns of the sub-aperture image
					subaperture.at<Vec3f>(j, 2 * i)[k] = (float)Raw.at<Vec3b>(Y, X)[k];
				}
			}
		}
	}

	interpolate_subaperture(subaperture, Center);

	return;
}

// creating sub-aperture images by directmethod, masks of sub-aperture images
void viewCreate::extract_viewandmask(const Mat& Raw, Mat* View, Mat* Mask)
{
	int n = 0;
	int X, Y;
	double x, y;

	int rcy_min = -3;
	int rcy_max = 2;
	int rcx_min = 4;
	int rcx_max = 9;

	if (FRAME == 25){
		rcy_min = -3;
		rcy_max = 2;
		rcx_min = 4;
		rcx_max = 9;
	}
	else if (FRAME == 49){
		rcy_min = -4;
		rcy_max = 3;
		rcx_min = 3;
		rcx_max = 10;
	}
	else if (FRAME == 81){
		rcy_min = -5;
		rcy_max = 4;
		rcx_min = 2;
		rcx_max = 11;
	}
	else if (FRAME == 121){
		rcy_min = -6;
		rcy_max = 5;
		rcx_min = 1;
		rcx_max = 12;
	}
	else if (FRAME == 169){
		rcy_min = -7;
		rcy_max = 6;
		rcx_min = 0;
		rcx_max = 13;
	}

	for (int s = rcy_min; s < rcy_max; s++){
		for (int t = rcx_min; t < rcx_max; t++){
			// expanding double in x direction
			Mat subaperture = Mat::zeros(view_y, 2 * view_x, CV_64FC3);
			Mat mask = Mat::ones(view_y, 2 * view_x, CV_8UC1) * 3;
			for (int j = 0; j < view_y; j = j + 2){
				for (int i = 0; i < view_x; i++){
					// pixel position of center viewpoint in each subimage
					y = offset_y + space_y * j + slope_y * i;
					x = offset_x + space_x * i + slope_x * j;

					// round the pixel position
					Y = cvRound(y) + s;
					X = cvRound(x) + t;
					if (Y >= 0 && Y < gray.rows && X >= 0 && X < gray.cols){
						if ((Y % 2) == 0){
							if ((X % 2) == 0){		// G
								mask.at<unsigned char>(j, 2 * i + 1) = 1;
								subaperture.at<Vec3d>(j, 2 * i + 1)[0] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i + 1)[1] = (double)Raw.at<unsigned char>(Y, X);
								subaperture.at<Vec3d>(j, 2 * i + 1)[2] = 0.0;
							}
							else{							// R
								mask.at<unsigned char>(j, 2 * i + 1) = 2;
								subaperture.at<Vec3d>(j, 2 * i + 1)[0] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i + 1)[1] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i + 1)[2] = (double)Raw.at<unsigned char>(Y, X);
							}
						}
						else{
							if ((X % 2) == 0){						// B
								mask.at<unsigned char>(j, 2 * i + 1) = 0;
								subaperture.at<Vec3d>(j, 2 * i + 1)[0] = (double)Raw.at<unsigned char>(Y, X);
								subaperture.at<Vec3d>(j, 2 * i + 1)[1] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i + 1)[2] = 0.0;
							}
							else{										// G
								mask.at<unsigned char>(j, 2 * i + 1) = 1;
								subaperture.at<Vec3d>(j, 2 * i + 1)[0] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i + 1)[1] = (double)Raw.at<unsigned char>(Y, X);
								subaperture.at<Vec3d>(j, 2 * i + 1)[2] = 0.0;
							}
						}
					}
				}
			}

			for (int j = 1; j < view_y; j = j + 2){
				for (int i = 0; i < view_x; i++){
					// pixel position of center viewpoint in each subimage
					y = offset_y + space_y * j + slope_y * i;
					x = offset_x - space_x / 2.0 + space_x * i + slope_x * j;

					// round the pixel position
					Y = cvRound(y) + s;
					X = cvRound(x) + t;
					if (Y >= 0 && Y < gray.rows - 1 && X >= 0 && X < gray.cols - 1){
						if ((Y % 2) == 0){
							if ((X % 2) == 0){		// G
								mask.at<unsigned char>(j, 2 * i) = 1;
								subaperture.at<Vec3d>(j, 2 * i)[0] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i)[1] = (double)Raw.at<unsigned char>(Y, X);
								subaperture.at<Vec3d>(j, 2 * i)[2] = 0.0;
							}
							else{							// R
								mask.at<unsigned char>(j, 2 * i) = 2;
								subaperture.at<Vec3d>(j, 2 * i)[0] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i)[1] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i)[2] = (double)Raw.at<unsigned char>(Y, X);
							}
						}
						else{
							if ((X % 2) == 0){						// B
								mask.at<unsigned char>(j, 2 * i) = 0;
								subaperture.at<Vec3d>(j, 2 * i)[0] = (double)Raw.at<unsigned char>(Y, X);
								subaperture.at<Vec3d>(j, 2 * i)[1] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i)[2] = 0.0;
							}
							else{										// G
								mask.at<unsigned char>(j, 2 * i) = 1;
								subaperture.at<Vec3d>(j, 2 * i)[0] = 0.0;
								subaperture.at<Vec3d>(j, 2 * i)[1] = (double)Raw.at<unsigned char>(Y, X);
								subaperture.at<Vec3d>(j, 2 * i)[2] = 0.0;
							}
						}
					}
				}
			}
			Mat intView;
			subaperture.convertTo(intView, CV_8UC3);
			intView.copyTo(View[n]);
			mask.copyTo(Mask[n]);
			subaperture.release();
			mask.release();
			n++;
		}
	}

	return;
}

void viewCreate::interpolate_subaperture(const Mat& Input, Mat& Output)
{
	int k;
	Input.copyTo(Output);
	Vec3f *outputPtr = (Vec3f *)(Output.ptr<Vec3f>(0));
	for (int j = 0; j < Input.rows; j++){
		for (int i = 0; i < Input.cols; i++){
			if ((j % 2) == 0 && (i % 2) == 0){
				if (i == 0){
					for (k = 0; k < 3; k++){
						(*outputPtr)[k] = Input.at<Vec3f>(j, 1)[k];
					}
				}
				else if (i == Input.cols - 1){
					for (k = 0; k < 3; k++){
						(*outputPtr)[k] = Input.at<Vec3f>(j, Input.cols - 2)[k];
					}
				}
				else{
					for (k = 0; k < 3; k++){
						(*outputPtr)[k] = (Input.at<Vec3f>(j, i - 1)[k] + Input.at<Vec3f>(j, i + 1)[k]) / 2.0f;
					}
				}
			}
			else if ((j % 2) == 1 && (i % 2) == 1){
				if (i != Input.cols - 1){
					for (k = 0; k < 3; k++){
						(*outputPtr)[k] = (Input.at<Vec3f>(j, i - 1)[k] + Input.at<Vec3f>(j, i + 1)[k]) / 2.0f;
					}
				}
			}
			outputPtr++;
		}
	}

	return;
}
